import unittest
import os
import sbsv

RESOURCE_DIR = os.path.join(os.path.dirname(__file__), "resources")


class TestNullable(unittest.TestCase):
    def test_nullable(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file?: str]")
        test_str = "[mem] [neg] id is [id myid] and file is [file]\n"
        test_str += "[mem] [neg] id is [id myid2] and file is [file [myfile2!]]\n"
        result = parser.loads(test_str)
        self.assertEqual(result["mem"]["neg"][0]["id"], "myid")
        self.assertEqual(result["mem"]["neg"][0]["file"], None)
        self.assertEqual(result["mem"]["neg"][1]["id"], "myid2")
        self.assertEqual(result["mem"]["neg"][1]["file"], "[myfile2!]")



ESCAPE_MAP = {
    "\b": "\\b",
    "\t": "\\t",
    "\n": "\\n",
    "\f": "\\f",
    "\r": "\\r",
    '"': '\\"',
    "/": "\\/",
    "\\": "\\\\",
    "[": "\\[",
    "]": "\\]",
    ",": "\\,",
    ":": "\\:",
}


def get_schema_id(*args) -> str:
    return "$".join(args)


def get_schema_name_list(schema: str) -> list:
    return schema.split("$")


def escape_str(s: str) -> str:
    escaped_str = s
    for raw, escaped in ESCAPE_MAP.items():
        escaped_str = escaped_str.replace(raw, escaped)
    return escaped_str


def unescape_str(s: str) -> str:
    unescaped_str = s
    for raw, escaped in ESCAPE_MAP.items():
        unescaped_str = unescaped_str.replace(escaped, raw)
    return unescaped_str



import unittest
import os
import sbsv

RESOURCE_DIR = os.path.join(os.path.dirname(__file__), "resources")


class TestParser(unittest.TestCase):
    def test_parser_remove(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file: str]")
        test_str = "[mem] [neg] id is [id myid] and file is [file myfile!]\n"
        test_str += "[mem] [neg] id is [id myid2] and file is [file [myfile2!]]\n"
        result = parser.loads(test_str)
        self.assertEqual(result["mem"]["neg"][0]["id"], "myid")
        self.assertEqual(result["mem"]["neg"][0]["file"], "myfile!")
        self.assertEqual(result["mem"]["neg"][1]["id"], "myid2")
        self.assertEqual(result["mem"]["neg"][1]["file"], "[myfile2!]")

    def test_parser_name_matching(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file: str]")
        parser.add_schema(
            "[mem] [both] [id$neg: str] [file$neg: str] [id$pos: str] [file$pos: str]"
        )
        test_str = "[mem] [neg] [id myid] [this is unknown] [file myfile!]\n"
        test_str += "[mem] [both] [id myid] [file myfile!] [id myid2] [file myfile2!] [at the end]\n"
        test_str += "[mem] [both] [id myid] [fake id] [file myfile!] [my-id hehe] [id myid3] [file myfile3!]\n"
        result = parser.loads(test_str)
        self.assertEqual(result["mem"]["neg"][0]["id"], "myid")
        self.assertEqual(result["mem"]["neg"][0]["file"], "myfile!")
        self.assertEqual(result["mem"]["both"][0]["id$neg"], "myid")
        self.assertEqual(result["mem"]["both"][0]["file$neg"], "myfile!")
        self.assertEqual(result["mem"]["both"][0]["id$pos"], "myid2")
        self.assertEqual(result["mem"]["both"][0]["file$pos"], "myfile2!")
        self.assertEqual(result["mem"]["both"][1]["id$neg"], "myid")
        self.assertEqual(result["mem"]["both"][1]["file$neg"], "myfile!")
        self.assertEqual(result["mem"]["both"][1]["id$pos"], "myid3")
        self.assertEqual(result["mem"]["both"][1]["file$pos"], "myfile3!")



import unittest
import os
import sbsv

RESOURCE_DIR = os.path.join(os.path.dirname(__file__), "resources")


class TestId(unittest.TestCase):
    def test_id(self):
        test_file = os.path.join(RESOURCE_DIR, "test_schema.sbsv")
        parser = sbsv.parser()
        parser.add_schema("[node] [id: int] [value: int]")
        parser.add_schema("[edge] [src: int] [dst: int] [value: int]")
        with open(test_file, "r") as f:
            result = parser.load(f)
        self.assertEqual(len(result["node"]), 4)
        self.assertEqual(len(result["edge"]), 5)
        elems = parser.get_result_in_order(["node", "edge"])
        expected_values = [
            {"id": 1, "value": 2},
            {"id": 2, "value": 3},
            {"src": 1, "dst": 2, "value": 6},
            {"src": 1, "dst": 3, "value": 10},
            {"id": 3, "value": 5},
            {"id": 4, "value": 7},
            {"src": 2, "dst": 3, "value": 15},
            {"src": 4, "dst": 5, "value": 35},
            {"src": 4, "dst": 1, "value": 14},
        ]
        for i, elem in enumerate(elems):
            self.assertEqual(elem.data, expected_values[i])

    def test_id_sub_schema(self):
        test_file = os.path.join(RESOURCE_DIR, "test_schema_sub.sbsv")
        parser = sbsv.parser()
        parser.add_schema("[graph] [node] [id: int] [value: int]")
        parser.add_schema("[graph] [edge] [src: int] [dst: int] [value: int]")
        with open(test_file, "r") as f:
            result = parser.load(f)
        elems = parser.get_result_in_order(
            [sbsv.get_schema_id("graph", "node"), sbsv.get_schema_id("graph", "edge")]
        )
        self.assertEqual(len(result["graph"]["node"]), 4)
        self.assertEqual(len(result["graph"]["edge"]), 5)
        self.assertEqual(len(elems), 9)
        self.assertEqual(elems[0].data, {"id": 1, "value": 2})
        self.assertEqual(elems[1].data, {"id": 2, "value": 3})
        self.assertEqual(elems[2].data, {"id": 3, "value": 5})
        self.assertEqual(elems[3].data, {"id": 4, "value": 7})
        self.assertEqual(elems[4].data, {"src": 1, "dst": 2, "value": 6})
        self.assertEqual(elems[5].data, {"src": 1, "dst": 3, "value": 10})
        self.assertEqual(elems[6].data, {"src": 2, "dst": 3, "value": 15})
        self.assertEqual(elems[7].data, {"src": 4, "dst": 5, "value": 35})
        self.assertEqual(elems[8].data, {"src": 4, "dst": 1, "value": 14})



import unittest
import os
import sbsv

RESOURCE_DIR = os.path.join(os.path.dirname(__file__), "resources")


class TestSchema(unittest.TestCase):
    def test_add_schema(self):
        test_file = os.path.join(RESOURCE_DIR, "test_schema.sbsv")
        parser = sbsv.parser()
        parser.add_schema("[node] [id: int] [value: int]")
        parser.add_schema("[edge] [src: int] [dst: int] [value: int]")
        with open(test_file, "r") as f:
            result = parser.load(f)
        self.assertEqual(len(result["node"]), 4)
        self.assertEqual(len(result["edge"]), 5)

    def test_sub_schema(self):
        test_file = os.path.join(RESOURCE_DIR, "test_schema_sub.sbsv")
        parser = sbsv.parser()
        parser.add_schema("[graph] [node] [id: int] [value: int]")
        parser.add_schema("[graph] [edge] [src: int] [dst: int] [value: int]")
        with open(test_file, "r") as f:
            result = parser.load(f)
        # self.assertEqual(len(result["graph$node"]), 4)
        # self.assertEqual(len(result["graph$edge"]), 5)
        # print(result)
        self.assertEqual(len(result["graph"]["node"]), 4)
        self.assertEqual(len(result["graph"]["edge"]), 5)

    def test_sub_schema_2(self):
        parser = sbsv.parser()
        parser.add_schema(
            "[pacfix] [mem] [neg] [seed: int] [id: int] [hash: int] [time: int] [file: str]"
        )
        parser.add_schema(
            "[pacfix] [mem] [pos] [seed: int] [id: int] [hash: int] [time: int] [file: str]"
        )
        parser.add_schema(
            "[moo] [save] [seed: int] [moo-id: int] [fault: int] [path: bool] [val: int] [file: str] [mut: str] [time: int]"
        )
        parser.add_schema(
            "[vertical] [save] [seed: int] [id: int] [dfg-path: int] [cov: int] [prox: int] [adj: float] [mut: str] [time: int]"
        )
        parser.add_schema("[vertical] [dry-run] [id: int] [dfg-path: int] [res: int]")
        parser.add_schema(
            "[vertical] [valuation] [seed: int] [dfg-path: int] [hash: int] [id: int] [persistent: bool] [time: int]"
        )

    def test_schema_duplicated(self):
        parser = sbsv.parser()
        parser.add_schema(
            "[node] [id$0: int] [value$0: int] [id$1: int] [value$1: int]"
        )
        result = parser.loads("[node] [id 1] [value 2] [id 3] [value 4]\n")
        self.assertEqual(result["node"][0]["id$0"], 1)
        self.assertEqual(result["node"][0]["value$0"], 2)
        self.assertEqual(result["node"][0]["id$1"], 3)
        self.assertEqual(result["node"][0]["value$1"], 4)



from typing import List, Dict, Tuple, Set, TextIO, Callable, Any
import queue
from .utils import get_schema_id, get_schema_name_list, escape_str, unescape_str


class lexer:
    def __init__(self):
        pass

    @staticmethod
    def update_token(result: List[str], current: str, should_replace: bool):
        current = current.strip()
        if should_replace:
            current = unescape_str(current)
        result.append(current)

    @staticmethod
    def tokenize(line: str) -> List[str]:
        result = list()
        level = 0
        current = ""
        escape = False
        should_replace = False
        for c in range(len(line)):
            char = line[c]
            if escape:
                escape = False
                current += "\\" + char
                continue
            if char == "\\":
                escape = True
                should_replace = True
                continue
            if char == "[":
                level += 1
                if level == 1:
                    if len(current.strip()) > 0:
                        lexer.update_token(result, current, should_replace)
                        should_replace = False
                    current = ""
                    continue
            elif char == "]":
                level -= 1
                if level == 0:
                    lexer.update_token(result, current, should_replace)
                    should_replace = False
                    current = ""
                    continue
            if level > 0:
                current += char
        return result

    @staticmethod
    def token_split(token: str, delimiter: str) -> Tuple[str, str]:
        tokens = token.split(delimiter, maxsplit=1)
        if len(tokens) == 0:
            return "", ""
        if len(tokens) == 1:
            return tokens[0].strip(), ""
        return tokens[0].strip(), tokens[1].strip()

    @staticmethod
    def token_split_default(token: str) -> Tuple[str, str]:
        return lexer.token_split(token, None)

    @staticmethod
    def token_split_schema(token: str) -> Tuple[str, str]:
        return lexer.token_split(token, ":")


class SbsvData:
    schema_name: str
    data: Dict[str, Any]
    id: int

    def __init__(self, schema_name: str, data: Dict[str, Any], id: int):
        self.schema_name = schema_name
        self.data = data
        self.id = id

    def __getitem__(self, key: str) -> Any:
        return self.data[key]

    def __setitem__(self, key: str, value: Any):
        self.data[key] = value

    def __delitem__(self, key: str):
        del self.data[key]

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __str__(self) -> str:
        return f"[schema {self.schema_name}] [id {self.id}] [data {self.data}]"

    def __repr__(self) -> str:
        return f"SbsvData({self.__str__()})"

    def get_id(self) -> int:
        return self.id

    def get_name(self) -> str:
        return self.schema_name


class SbsvDataType:
    name: str
    name_with_tag: str
    type: str
    nullable: bool
    converter: Callable[[str], Any]
    sub_type: List["SbsvDataType"]

    def __init__(self, name_with_tag: str, type: str):
        self.nullable = name_with_tag.endswith("?")
        if self.nullable:
            name_with_tag = name_with_tag[:-1]
        self.name_with_tag = name_with_tag
        if "$" in self.name_with_tag:
            tokens = self.name_with_tag.split("$")
            self.name = tokens[0]
        else:
            self.name = name_with_tag
        self.type = type
        self.converter = self.add_converter(type)
        self.sub_type = list()

    @staticmethod
    def to_bool(value: str) -> bool:
        value = value.strip().lower()
        if value in ["t", "true", "y", "yes", "1"]:
            return True
        elif value in ["f", "false", "n", "no", "0"]:
            return False
        raise ValueError(f"Invalid boolean value: {value}")

    def add_converter(self, type: str) -> Callable[[str], Any]:
        # Primitive types
        if type == "int":
            return int
        if type == "float":
            return float
        if type == "str":
            return str
        if type == "bool":
            return SbsvDataType.to_bool
        # Complex types
        if type.startswith("list"):
            sub_type = type[5:-1]
            return lambda x: [
                SbsvDataType(sub_type, sub_type).convert(v) for v in lexer.tokenize(x)
            ]
        # Unsupported types
        return None

    def convert(self, value: str) -> Any:
        if value == "" and self.nullable:
            return None
        if self.converter is not None:
            return self.converter(value)
        return value

    def key(self) -> str:
        return self.name_with_tag

    def check_nullable(self) -> bool:
        return self.nullable

    def check_name(self, name: str) -> bool:
        return self.name == name


class Schema:
    original: str
    name: str
    schema: List[SbsvDataType]
    data: List[SbsvData]

    def __init__(self, s: str):
        self.original = s
        self.name = ""
        self.schema = list()
        self.data = list()
        tokens = lexer.tokenize(s)
        if len(tokens) <= 1:
            raise ValueError("Invalid schema: too short")
        self.name = tokens[0]
        for i in range(1, len(tokens)):
            key, value = lexer.token_split_default(tokens[i])
            if key == "":
                raise ValueError("Invalid schema: empty name")
            # Sub schema
            if value == "":
                self.name = f"{self.name}${key}"
                continue
            key, value = lexer.token_split_schema(tokens[i])
            # Normal schema
            self.schema.append(SbsvDataType(key, value))

    @staticmethod
    def preprocess(line: str) -> Tuple[str, List[str]]:
        tokens = lexer.tokenize(line)
        name: str = None
        data: List[str] = list()
        may_have_sub_schema = True
        for i in range(len(tokens)):
            key, value = lexer.token_split_default(tokens[i])
            if key != "" and value == "" and may_have_sub_schema:
                if name is None:
                    name = key
                else:
                    name = f"{name}${key}"
            else:
                may_have_sub_schema = False
                data.append(tokens[i])
        return name, data

    def parse(self, tokens: List[str]) -> Dict[str, Any]:
        result = dict()
        if len(tokens) < len(self.schema):
            raise ValueError("Invalid data: too short")
        q = queue.Queue(len(tokens))
        for token in tokens:
            q.put(token)
        for schema_type in self.schema:
            done = False
            while not q.empty():
                elem = q.get()
                key, value = lexer.token_split_default(elem)
                if key == "":
                    raise ValueError("Invalid data: empty name")
                if not schema_type.check_name(key):
                    continue
                if value == "" and not schema_type.check_nullable():
                    raise ValueError("Invalid data: empty value")
                result[schema_type.key()] = schema_type.convert(value)
                done = True
                break
            if not done:
                raise ValueError("Invalid data: missing key")
        return result

    def get_data(self) -> List[SbsvData]:
        return self.data

    def append_data(self, data: SbsvData):
        self.data.append(data)


class parser:
    data: List[SbsvData]
    result: dict
    schema: Dict[str, Schema]
    ignore_unknown: bool

    def __init__(self, ignore_unknown: bool = True):
        self.data = list()
        self.result = dict()
        self.schema = dict()
        self.ignore_unknown = ignore_unknown

    # New parser with same schema
    def clone(self) -> "parser":
        result = parser(self.ignore_unknown)
        result.schema = self.schema.copy()
        return result

    def get_global_id(self) -> int:
        return len(self.data)

    def match_schema(self, line: str) -> Tuple[Schema, List[str]]:
        name, value = Schema.preprocess(line)
        if name not in self.schema:
            if self.ignore_unknown:
                return None, None
            raise ValueError(f"Schema not found: {name}")
        return self.schema[name], value

    def add_schema(self, schema: str):
        # 1. tokenize
        sc = Schema(schema)
        self.schema[sc.name] = sc

    def post_process(self):
        for key in self.schema:
            if "$" in key:
                tokens = key.split("$")
                tmp_root = self.result
                final_token = tokens[-1]
                for i in range(len(tokens)):
                    if i == len(tokens) - 1:
                        break
                    if tokens[i] not in tmp_root:
                        tmp_root[tokens[i]] = dict()
                    tmp_root = tmp_root[tokens[i]]
                tmp_root[final_token] = self.schema[key].get_data()
            else:
                self.result[key] = self.schema[key].get_data()

    def append_row_to_data(self, schema: Schema, row: Dict[str, Any]):
        sbsv_data = SbsvData(schema.name, row, self.get_global_id())
        self.data.append(sbsv_data)
        schema.append_data(sbsv_data)

    def parse_line(self, line: str):
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            return
        sc, tokens = self.match_schema(line)
        if sc is None:
            return
        row = sc.parse(tokens)
        self.append_row_to_data(sc, row)

    def load(self, fp: TextIO) -> dict:
        for line in fp:
            self.parse_line(line)
        self.post_process()
        return self.result

    def loads(self, s: str) -> dict:
        for line in s.split("\n"):
            self.parse_line(line)
        self.post_process()
        return self.result

    def get_result(self) -> dict:
        return self.result

    def get_result_in_order(self, schemas: List[str] = None) -> List[SbsvData]:
        if schemas is None:
            return self.data
        pq = queue.PriorityQueue()
        for schema in schemas:
            if schema not in self.schema:
                raise ValueError(f"Schema not found: {schema}")
            cur_schema = self.schema[schema]
            if len(cur_schema.get_data()) == 0:
                continue
            pq.put((cur_schema.get_data()[0].get_id(), cur_schema, 0))
        result = list()
        while not pq.empty():
            value, cur_schema, elem = pq.get()
            result.append(cur_schema.get_data()[elem])
            if elem < len(cur_schema.get_data()) - 1:
                next_value = cur_schema.get_data()[elem + 1].get_id()
                pq.put((next_value, cur_schema, elem + 1))
        return result



import unittest
import os
import sbsv

RESOURCE_DIR = os.path.join(os.path.dirname(__file__), "resources")


class TestLexer(unittest.TestCase):
    def test_lexer(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file: str]")
        parser.add_schema("[mem] [pos] [seed: int] [id: str] [file: str]")
        test_str = "[mem] [neg] [id str can have spaces] [file /path/to/file]\n"
        result = parser.loads(test_str)
        self.assertEqual(result["mem"]["neg"][0]["id"], "str can have spaces")

    def test_lexer_escape(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file: str]")
        parser.add_schema("[mem] [pos] [seed: int] [id: str] [file: str]")
        test_str = "[mem] [neg] [id should escape \\] this] [file /path/to/file]\n"
        test_str += '[mem] [pos] [seed 123] [id should escape \\]\\]\\ this] [file /path/to/file\\"]\n'
        result = parser.loads(test_str)
        self.assertEqual(result["mem"]["neg"][0]["id"], "should escape ] this")
        self.assertEqual(result["mem"]["pos"][0]["id"], "should escape ]]\\ this")
        self.assertEqual(result["mem"]["pos"][0]["file"], '/path/to/file"')

    def test_lexer_escape_file(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file: str]")
        parser.add_schema("[mem] [pos] [seed: int] [id: str] [file: str]")
        with open(os.path.join(RESOURCE_DIR, "test_lexer_escape.sbsv"), "r") as f:
            result = parser.load(f)
        self.assertEqual(result["mem"]["neg"][0]["id"], "should escape ] this")
        self.assertEqual(result["mem"]["pos"][0]["id"], "should escape ]]\\ this")
        self.assertEqual(result["mem"]["pos"][0]["file"], '/path/to/file"')

    def test_lexer_remove(self):
        parser = sbsv.parser()
        parser.add_schema("[mem] [neg] [id: str] [file: str]")
        test_str = "[mem] [neg] id is [id myid] and file is [file myfile!]\n"
        result = parser.loads(test_str)
        self.assertEqual(result["mem"]["neg"][0]["id"], "myid")
        self.assertEqual(result["mem"]["neg"][0]["file"], "myfile!")

    def test_escpae(self):
        self.assertEqual(sbsv.escape_str("this is a test"), "this is a test")
        self.assertEqual(sbsv.escape_str("this is a test]"), "this is a test\\]")
        self.assertEqual(sbsv.escape_str("this is a test\\"), "this is a test\\\\")
        self.assertEqual(sbsv.escape_str(",|]][[]]"), "\\,|\\]\\]\\[\\[\\]\\]")
        self.assertEqual(sbsv.unescape_str("this is a test"), "this is a test")
        self.assertEqual(sbsv.unescape_str("this is a test\\]"), "this is a test]")
        self.assertEqual(sbsv.unescape_str("this is a test\\\\"), "this is a test\\")
        self.assertEqual(sbsv.unescape_str("\\,|\\]\\]\\[\\[\\]"), ",|]][[]")
        self.assertEqual(
            sbsv.unescape_str(sbsv.escape_str("should escape ]]\\ this")),
            "should escape ]]\\ this",
        )
